/* dzl-list-box-row.h
 *
 * Copyright (C) 2017 Christian Hergert <chergert@redhat.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef DZL_LIST_BOX_ROW_H
#define DZL_LIST_BOX_ROW_H

#include <gtk/gtk.h>

G_BEGIN_DECLS

#define DZL_TYPE_LIST_BOX_ROW (dzl_list_box_row_get_type())

DZL_AVAILABLE_IN_ALL
G_DECLARE_DERIVABLE_TYPE (DzlListBoxRow, dzl_list_box_row, DZL, LIST_BOX_ROW, GtkListBoxRow)

struct _DzlListBoxRowClass
{
  GtkListBoxRowClass parent_class;
};

DZL_AVAILABLE_IN_ALL
GtkWidget *dzl_list_box_row_new (void);

G_END_DECLS

#endif /* DZL_LIST_BOX_ROW_H */
